﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;

namespace GalaxyAnalysis
{
    public partial class Form1 : Form
    {
        public Form1()
        {
            InitializeComponent();
            UpdateBrightnessTB();
        }

        private void reduceSizeCheck_CheckedChanged(object sender, EventArgs e)
        {
            if (reduceSizeCheck.Checked)
            {
                imagePreview.Dock = DockStyle.Fill;
                imagePreview.SizeMode = PictureBoxSizeMode.StretchImage;
            }
            else
            {
                imagePreview.Dock = DockStyle.None;
                imagePreview.SizeMode = PictureBoxSizeMode.AutoSize;
            }
        }

        Bitmap bmp = null;

        SortedList<Rectangle, List<Point>> detections = new SortedList<Rectangle, List<Point>>();

        private void loadImage_Click(object sender, EventArgs e)
        {
            OpenFileDialog dialog = new OpenFileDialog();
            if (dialog.ShowDialog() == DialogResult.OK)
            {
                detections = new SortedList<Rectangle, List<Point>>();
                detectedGalaxies.Items.Clear();
                bmp = (Bitmap)Bitmap.FromFile(dialog.FileName);
                UpdateImage();

            }
        }

        private void UpdateImage()
        {
            if (bmp == null)
            {
                return;
            }

            imagePreview.Image = RemoveBackground(bmp);
        }

        private Image RemoveBackground(Bitmap bmp)
        {
            Bitmap removedBackground = new Bitmap(bmp);
            for (int i = 0; i < removedBackground.Width; i++)
            {
                for (int j = 0; j < removedBackground.Height; j++)
                {
                    Color c = removedBackground.GetPixel(i, j);
                    if (IsBackground(c, backBrightness.Value))
                    {
                        removedBackground.SetPixel(i, j, Color.Black);
                    }

                }
            }

 

            if (detectedGalaxies.SelectedIndices.Count > 0)
            {
                foreach (int index in detectedGalaxies.SelectedIndices)
                {
                    List<Point> selectedPoints = detections.Values[index];
                    foreach (Point point in selectedPoints)
                    {
                        removedBackground.SetPixel(point.X, point.Y, Color.Green);
                    }
                    
                }
            }

            using(Graphics g = Graphics.FromImage(removedBackground))
            {
                foreach (Rectangle r in detections.Keys)
                {
                    g.DrawRectangle(Pens.Red, r);
                }

                if (detectedGalaxies.SelectedIndices.Count > 0)
                {
                    foreach (int index in detectedGalaxies.SelectedIndices)
                    {
                        Rectangle selected = detections.Keys[index];
                        g.DrawRectangle(Pens.Cyan, selected);

                    }
                }

            }

 

            return removedBackground;

        }

        private bool IsBackground(Color c, int p)
        {

            if(GetBRAverage(c)>p)
            {
                return false;
            }

            return true;

        }




        private void backBrightness_Scroll(object sender, EventArgs e)
        {
            UpdateImage();
        }

        private void runAnalysis_Click(object sender, EventArgs e)
        {
            Bitmap findWhere = (Bitmap)imagePreview.Image;
            detections = FindGalaxies(findWhere);
            listUpdate = true;
            detectedGalaxies.Items.Clear();
            int index = 1;
            foreach (Rectangle r in detections.Keys)
            {
                decimal brightness = CalculateBrightness(detections[r], findWhere);
                string resultString = MakeResultString(brightness, r, detections[r].Count, index);
                detectedGalaxies.Items.Add(resultString);
                index++;
            }
            UpdateImage();
            listUpdate = false;
        }

        private string MakeResultString(decimal brightness, Rectangle r, int count, int index)
        {
            return string.Format("Nr:\t{6}\tLocation:\t{0}\t{1}\tSize:\t{2}\t{3}\tPointsCount:\t{4}\tBrightness:\t{5}", r.X, r.Y, r.Width, r.Height, count, brightness.ToString("0.00"), index);
        }

        private decimal CalculateBrightness(List<Point> list, Bitmap findWhere)
        {
            
            decimal current = 0;
            foreach (Point p in list)
            {
                Color point = findWhere.GetPixel(p.X, p.Y);
                decimal brAvegage = GetBRAverage( point);
                brAvegage = brAvegage/(decimal)255;
                current += brAvegage;
            }

            return current / (decimal)list.Count;

        }

        private static decimal GetBRAverage( Color point)
        {
            decimal brAvegage = (point.R + point.G + point.B) / (decimal)(3);
            return brAvegage;
        }


        class RectCompared : IComparer<Rectangle>
        {
            #region IComparer<Rectangle> Members

            public int Compare(Rectangle x, Rectangle y)
            {
                int rez = x.Left.CompareTo(y.Left);
                if (rez != 0)
                {
                    return rez;
                }

                rez = x.Right.CompareTo(y.Right);
                if (rez != 0)
                {
                    return rez;
                }
                rez = x.Top.CompareTo(y.Top);
                if (rez != 0)
                {
                    return rez;
                }

                rez = x.Bottom.CompareTo(y.Bottom);
                if (rez != 0)
                {
                    return rez;
                }

                return 0;
            }

            #endregion
        }


        public class BitmapGraph
        {
            class PointCompared : IComparer<Point>
            {

                #region IComparer<Point> Members

                public int Compare(Point x, Point y)
                {
                    int rez = x.X.CompareTo(y.X);
                    if (rez != 0)
                    {
                        return rez;
                    }

                    return x.Y.CompareTo(y.Y);


                }

                #endregion
            }

            SortedDictionary<Point, List<Point>> graphForward = new SortedDictionary<Point, List<Point>>(new PointCompared());
            SortedDictionary<Point, List<Point>> graphBackward = new SortedDictionary<Point, List<Point>>(new PointCompared());

            public void AddEdge(Point from, Point to)
            {
                AddPoint(graphForward, from, to);
                AddPoint(graphBackward, to, from);
             }

            private void AddPoint(SortedDictionary<Point, List<Point>> graphForward, Point from, Point to)
            {
                if (!graphForward.ContainsKey(from))
                {
                    graphForward[from] = new List<Point>();
                }
                graphForward[from].Add(to);
            }

            public bool Contains(Point from, Point to)
            {
                if (!graphForward.ContainsKey(from))
                {
                    return false;
                }

                if (graphBackward.ContainsKey(from))
                {
                    return false;
                }

                foreach(Point check in graphForward[from])
                {
                    if(check==to)
                    {
                        return true;
                    }
                }

                return false;
            }


            internal bool HasPoints()
            {
                if (graphForward.Count > 0)
                {
                    return true;
                }
                return false;
            }

            internal Point GetFirstPoint()
            {
                foreach (Point pt in graphForward.Keys)
                {
                    return pt;
                }
                throw new Exception();
            }

            internal List<Point> FindAllConnected(Point pt)
            {
                SortedDictionary<Point, List<Point>> visited = new SortedDictionary<Point, List<Point>>(new PointCompared());
                visited.Add(pt, new List<Point>());
                VisitAll(pt,visited);
                return new List<Point>(visited.Keys);
            }

            private void VisitAll(Point pt, SortedDictionary<Point, List<Point>> visited)
            {
                //Dictionary<Point, int> visitDepth = new Dictionary<Point, int>();
                //visitDepth.Add(pt, 0);
                List<Point> depth = new List<Point>();
                bool hasVisitedNew = true;
                Point current = pt;
                depth.Add(current);
                while (depth.Count>0)
                {
                    //Point last = visitDepth[visitDepth.Count - 1];

                    VisitAll(current, visited, graphForward);
                    VisitAll(current, visited, graphBackward);

                    hasVisitedNew = false;
                    foreach (Point visit in visited[current])
                    {
                        if (!visited.ContainsKey(visit))
                        {
                            visited.Add(visit, new List<Point>());
                            depth.Add(visit);
                            current = visit;
                            hasVisitedNew = true;
                            break;
                        }
                    }

                    if (!hasVisitedNew)
                    {
                        depth.Remove(current);
                        if (depth.Count > 0)
                        {
                            current = depth[depth.Count - 1];
                        }
                        else
                        {
                            current = Point.Empty;
                        }
                    }


                }


            }

            private bool VisitAll( Point pt, SortedDictionary<Point, List<Point>> visited,  SortedDictionary<Point, List<Point>> graphForward)
            {
                bool visitedNew = false;
                foreach (Point check in Enumerate(graphForward, pt))
                {
                    if (!visited[pt].Contains(check))
                    {
                        if (!visited[pt].Contains(check))
                        {
                            visited[pt].Add(check);
                            //visited.Add(check, new List<Point>());
                            //visited[pt].Add(check);
                            visitedNew = true;
                        }
                    }
                }
                return visitedNew;
            }

            private IEnumerable<Point> Enumerate(SortedDictionary<Point, List<Point>> check, Point pt)
            {
                if (check.ContainsKey(pt))
                {
                    foreach (Point p in check[pt])
                    {
                        yield return p;
                    }

                }
            }

            /// <summary>
            /// Does not remove to point references, only from,
            /// but should not matter as all the rest is removed by using this method
            /// </summary>
            /// <param name="connPt"></param>
            internal void RemovePoint(Point connPt)
            {
                if (graphForward.ContainsKey(connPt))
                {
                    graphForward.Remove(connPt);
                }

                if (graphBackward.ContainsKey(connPt))
                {
                    graphBackward.Remove(connPt);
                }
            }
        }




        private SortedList<Rectangle, List<Point>> FindGalaxies(Bitmap analyzedBitmap)
        {
            SortedList<Rectangle, List<Point>> galaxies = new SortedList<Rectangle, List<Point>>(new RectCompared());

            BitmapGraph graph = new BitmapGraph();
            for (int x = 0; x < analyzedBitmap.Width; x++)
            {
                for (int y = 0; y < analyzedBitmap.Height; y++)
                {
                    Color pixelColor = analyzedBitmap.GetPixel(x, y);
                    if (!IsBackground(pixelColor, backBrightness.Value))
                    {
                        Point ptFrom = new Point(x, y);

                        foreach (Point ptTo in MakeNeighborPoints(ptFrom, analyzedBitmap))
                        {
                            if (!graph.Contains(ptFrom, ptTo))
                            {
                                graph.AddEdge(ptFrom, ptTo);
                            }
                            if (!graph.Contains(ptTo, ptFrom))
                            {
                                graph.AddEdge(ptTo, ptFrom);
                            }

                        }
                        
                    }
                }
            }

            while (graph.HasPoints())
            {
                Point pt = graph.GetFirstPoint();
                List<Point> connected = graph.FindAllConnected(pt);
                foreach (Point connPt in connected)
                {
                    graph.RemovePoint(connPt);
                }

                Rectangle rect = MakeRectangle(connected);
                galaxies.Add(rect, connected);

            }


            return galaxies;
        }

        private IEnumerable<Point> MakeNeighborPoints(Point ptFrom, Bitmap analyzedBitmap)
        {
            for (int x = -1; x < 2; x++)
            {
                for (int y = -1; y < 2; y++)
                {
                    if (x == 0)
                    {
                        if (y == 0)
                        {
                            continue;
                        }
                    }
                    Point directionNew = new Point(x, y);
                    Point checkPoint = Add(ptFrom, directionNew);

                    if (IsPointValid(checkPoint, analyzedBitmap))
                    {
                        yield return checkPoint;
                    }

                }
            }
        }



        private bool IsContainedIn(SortedList<Rectangle, List<Point>> galaxies, Point pt)
        {
            foreach (Rectangle r in galaxies.Keys)
            {
                if (Contains(r,pt))
                {
                    foreach (Point check in galaxies[r])
                    {
                        //Rectangle rr = MakeRectangle(galaxies[r]);

                        if (check == pt)
                        {
                            return true;
                        }
                    }
                }
            }
            return false;
        }

        private bool Contains(Rectangle r, Point pt)
        {
            if (pt.X < r.X)
            {
                return false;
            }

            if (pt.Y < r.Y)
            {
                return false;
            }

            if (pt.X > r.Right)
            {
                return false;
            }

            if (pt.Y > r.Bottom)
            {
                return false;
            }

            return true;

        }

        private Rectangle MakeRectangle(List<Point> allAttached)
        {
            int minx = int.MaxValue;
            int miny = int.MaxValue;
            int maxx = int.MinValue;
            int maxy = int.MinValue;

            foreach (Point pt in allAttached)
            {
                minx = Math.Min(minx, pt.X);
                maxx = Math.Max(maxx, pt.X);

                miny = Math.Min(miny, pt.Y);
                maxy = Math.Max(maxy, pt.Y);

            }

            return new Rectangle(minx, miny, maxx - minx, maxy - miny);

        }


        private Point Add(Point current, Point direction)
        {
            return new Point(current.X + direction.X, current.Y + direction.Y);
            //throw new NotImplementedException();
        }

        private bool IsPointValid(Point current, Bitmap analyzedBitmap)
        {
            if (current.X < 0)
            {
                return false;
            }

            if (current.Y < 0)
            {
                return false;
            }

            if (current.X >= analyzedBitmap.Width)
            {
                return false;
            }

            if (current.Y >= analyzedBitmap.Height)
            {
                return false;
            }

            if (IsBackground(analyzedBitmap.GetPixel(current.X, current.Y), backBrightness.Value))
            {
                return false;
            }


            return true;
        }

        private void copyResults_Click(object sender, EventArgs e)
        {
            string total = "";
            foreach (object o in detectedGalaxies.Items)
            {
                total += string.Format("{0}\r\n", o.ToString());
            }
            Clipboard.SetText(total);
        }

        private void backBrightness_ValueChanged(object sender, EventArgs e)
        {

        }

        private void backBrightness_Scroll_1(object sender, EventArgs e)
        {
            UpdateBrightnessTB();

        }

        private void UpdateBrightnessTB()
        {
            changingBrightness = true;
            brightnessLevelText.Text = backBrightness.Value.ToString();
            changingBrightness = false;
        }
        bool changingBrightness = false;
        private void brightnessLevelText_TextChanged(object sender, EventArgs e)
        {
            if (changingBrightness)
            {
                return;

            }

            int parsed;
            if (int.TryParse(brightnessLevelText.Text, out parsed))
            {
                if (IsValidBrighnessValue(parsed))
                {
                    backBrightness.Value = (int)parsed;
                }
            }

        }

        private bool IsValidBrighnessValue(int parsed)
        {
            if (parsed < backBrightness.Minimum)
            {
                return false;
            }

            if (parsed > backBrightness.Maximum)
            {
                return false;
            }
            return true;
        }

        private void refreshPicture_Click(object sender, EventArgs e)
        {
            UpdateImage();
        }


        bool listUpdate = false;
        private void detectedGalaxies_SelectedIndexChanged(object sender, EventArgs e)
        {
            if (listUpdate)
            {
                return;
            }

            UpdateImage();
        }


    }
}
